using System;
using System.Linq;
using System.Collections.Generic;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using HabitSharp.Extensions;
using HabitSharp.Notifications;

/// <summary>
/// A converter to create subclasses of Notification when possible to fetch more information.
/// If a notification is of an unknown type, will create a simple Notification class.
/// </summary>
/// <remarks>Serialization to JSON is not supported; this converter can only perform deserialization.</remarks>
public class NotificationConverter : JsonConverter
{
    /// <summary>
    /// Get a class derived from Notification is there is one matching the given notification type, or return Notification if none is found.
    /// </summary>
    /// <param name="notificationType">The notification type string from JSON data.</param>
    /// <returns>System.Type for the Notification class or one of its derived classes</returns>
    public Type FindNotificationSubclass(string notificationType) {
        // Find all Notification subclasses in the same assembly
        return typeof(Notification).Assembly.GetTypes()
            .Where(t => t.IsSubclassOf(typeof(Notification)))
            .SingleOrDefault(
                // Find the only one subclass that has a NotificationType attribute with the right type string
                t => Attribute.GetCustomAttributes(t, typeof(NotificationTypeAttribute)).Any(
                    attr => ((NotificationTypeAttribute)attr).TypeString.Equals(notificationType, StringComparison.InvariantCultureIgnoreCase)
                )
            ) ?? typeof(Notification); // Default to the Notification class
    }

    public override bool CanConvert(Type objectType) => objectType.IsOrSubclass(typeof(Notification));

    public override object ReadJson(JsonReader reader, Type objectType, object existingValue, JsonSerializer serializer)
    {
        JObject jo = JObject.Load(reader);
        Type noteClass = FindNotificationSubclass((string)jo["type"]);
        return Convert.ChangeType(jo.ToObject(noteClass), noteClass);
    }

    public override bool CanWrite => false;

    public override void WriteJson(JsonWriter writer, object value, JsonSerializer serializer) => throw new NotImplementedException();
}
